
#!/usr/bin/env python3
"""
DiT-based Block Type Prediction
================================

This script replaces the original “predict‑next‑coordinate” task with a new task:
given the *coordinates* of every block in a Minecraft‑style building, infer the
*block ID* (material) of each block.

Key design points
-----------------
* **DiT backbone** –  we use a minimal 3‑layer Transformer encoder which we dub
  “DiT‑Mini”.  Every coordinate token is embedded and processed in parallel.
* **PyTorch Lightning** – training and validation loops are implemented with a
  single ``LightningModule`` so that you get automatic logging, checkpointing
  and GPU support.
* **Single file** – everything (dataset, model, training script, synthetic test
  scene) lives here so you can just:

    ``python dit_blocktypes.py --data_dir path/to/3d_craft --batch_size 128``

The model is saved to ``logs/`` after every epoch.  During training, each **step**
prints the *instantaneous* loss and the *smoothed* moving‑average loss, and at
the end of every **epoch** we run inference on a tiny “matchbox” house that has
a few known IDs and many ``‑1`` placeholders.  The predicted IDs are printed as
an ASCII map so you can eyeball whether the model has learned context.

Command‑line arguments
----------------------
* ``--data_dir``  Folder that contains the original 3D‑Craft ``houses`` data.
* ``--batch_size`` Training batch size (default 64).
* ``--epochs``     Number of training epochs (default 10).
* ``--lr``         Learning rate (default 1e-3).

Dependencies
~~~~~~~~~~~~
* PyTorch ≥ 2.0
* PyTorch Lightning ≥ 2.2
* NumPy

Install Lightning quickly with:

    ``pip install torch torchvision torchaudio pytorch-lightning``

"""

import argparse
import math
import os
from pathlib import Path
from typing import List, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import lightning as L
from torch.utils.data import DataLoader, Dataset

# --------------------------------------------------------------------------------
#  Dataset
# --------------------------------------------------------------------------------
class Craft3DBlockTypeDataset(Dataset):
    """
    Flattens the original Craft3D "house" annotations into (coord → type) pairs.

    Each item is a tuple:
        coords : (3,) int tensor  – absolute (x,y,z) coordinates
        label  : ()   int tensor  – block ID in [0,255]

    We **discard** the temporal order – we only care about static structure.
    """

    NUM_TYPES = 256

    def __init__(self, data_dir: str, split: str = "train"):
        super().__init__()
        self.data_dir = Path(data_dir)
        splits_path = self.data_dir / "splits.json"
        if not splits_path.exists():
            raise FileNotFoundError(f"Cannot find {splits_path}. "
                                    "Download the 3D‑Craft dataset first.")
        with splits_path.open() as f:
            split_files = json.load(f)[split]

        self.coords: List[torch.Tensor] = []
        self.labels: List[int] = []

        for fname in split_files:
            placed = Path(self.data_dir / "houses" / fname / "placed.json")
            if not placed.exists():
                continue
            with placed.open() as f:
                traj = json.load(f)

            final_blocks = {}
            for tstamp, _, coord, block_info, action in traj:
                key = tuple(coord)
                if action == "B":  # break → remove block
                    final_blocks.pop(key, None)
                else:              # place
                    final_blocks[key] = int(block_info[0])

            for (x, y, z), btype in final_blocks.items():
                self.coords.append(torch.tensor([x, y, z], dtype=torch.float32))
                self.labels.append(btype)

        self.coords = torch.stack(self.coords)
        self.labels = torch.tensor(self.labels, dtype=torch.long)

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        return self.coords[idx], self.labels[idx]

# --------------------------------------------------------------------------------
#  Model – “DiT‑Mini”
# --------------------------------------------------------------------------------
class CoordEmbedding(nn.Module):
    """Embed (x,y,z) absolute coordinates with sinusoidal encoding + linear proj."""

    def __init__(self, d_model: int = 128):
        super().__init__()
        self.lin = nn.Linear(3 * 2 * 10, d_model, bias=False)  # 10 freq bands

    def forward(self, coords: torch.Tensor):
        # coords: (B,3)
        freqs = 2 ** torch.arange(10, device=coords.device, dtype=torch.float32)
        coords = coords.unsqueeze(-1) * freqs  # (B,3,10)
        coords = torch.cat([coords.sin(), coords.cos()], dim=-1).flatten(1)
        return self.lin(coords)


class DiTMini(nn.Module):
    """Very small Transformer classifier."""

    def __init__(self, num_types: int = 256, d_model: int = 128, nhead: int = 8, nlayers: int = 3):
        super().__init__()
        self.embed = CoordEmbedding(d_model)
        encoder_layer = nn.TransformerEncoderLayer(d_model=d_model, nhead=nhead, batch_first=True)
        self.backbone = nn.TransformerEncoder(encoder_layer, num_layers=nlayers)
        self.cls = nn.Linear(d_model, num_types)

    def forward(self, coords: torch.Tensor):
        """
        coords : (B,3) float – absolute positions
        returns: (B,C) logits
        """
        x = self.embed(coords)
        # Transformer expects a sequence; we use length 1 per sample.
        x = x.unsqueeze(1)           # (B,1,d_model)
        x = self.backbone(x)[:, 0]   # (B,d_model)
        return self.cls(x)


# --------------------------------------------------------------------------------
#  Lightning module
# --------------------------------------------------------------------------------
class BlockTypeModule(L.LightningModule):
    def __init__(self, hparams):
        super().__init__()
        self.save_hyperparameters(hparams)
        self.model = DiTMini()
        self.criterion = nn.CrossEntropyLoss()
        self.train_ma = 0.0
        self.example_input_array = torch.zeros(4, 3)  # for Lightning summary

    def forward(self, coords):
        return self.model(coords)

    def training_step(self, batch, batch_idx):
        coords, labels = batch
        logits = self(coords)
        loss = self.criterion(logits, labels)
        # moving average
        self.train_ma = 0.98 * self.train_ma + 0.02 * loss.item()
        self.log("loss", loss, on_step=True, prog_bar=True)
        self.log("loss_ma", self.train_ma, on_step=True, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        coords, labels = batch
        logits = self(coords)
        loss = self.criterion(logits, labels)
        acc = (logits.argmax(dim=1) == labels).float().mean()
        self.log_dict({"val_loss": loss, "val_acc": acc}, prog_bar=True)

    def configure_optimizers(self):
        optim = torch.optim.AdamW(self.parameters(), lr=self.hparams.lr)
        sched = torch.optim.lr_scheduler.CosineAnnealingLR(optim, T_max=self.hparams.epochs)
        return {"optimizer": optim, "lr_scheduler": sched}

    # -------------------------------------------------------------------------
    #  Synthetic “matchbox house” demo
    # -------------------------------------------------------------------------
    @torch.no_grad()
    def on_train_epoch_end(self):
        demo_coords, demo_types = generate_matchbox_house()
        # Mask: set unknowns to -1
        mask = demo_types == -1
        preds = self.model(demo_coords[mask]).argmax(dim=1)
        demo_types[mask] = preds.cpu()

        # Pretty‑print an xy slice
        grid = {}
        for (x, y, z), t in zip(demo_coords.int().tolist(), demo_types.tolist()):
            grid.setdefault(z, {})[(x, y)] = t
        z0 = sorted(grid)[len(grid)//2]
        print(f"\n=== Demo slice z={z0} after epoch {self.current_epoch} ===")
        for y in range(6):
            row = []
            for x in range(6):
                t = grid[z0].get((x, y), -1)
                row.append(f"{t:3d}" if t >= 0 else "  .")
            print(" ".join(row))
        print("============================================\n")


# --------------------------------------------------------------------------------
#  Utility – generate a tiny rectangular house (“matchbox”)
# --------------------------------------------------------------------------------
def generate_matchbox_house() -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Returns coords  (N,3) int32  and types (N,) int64 where a few blocks have an ID
    and the rest are −1.  Coordinates range 0‑5 in each axis so you can visualise.
    Block‑ID convention: 1 stone, 2 wood, 3 glass, 4 door.
    """
    coords = []
    types = []
    for x in range(6):
        for y in range(6):
            for z in [0, 5]:                 # floor & ceiling
                coords.append((x, y, z))
                types.append(1)              # stone
    for x in range(6):
        for z in range(1, 5):
            for y in [0, 5]:
                coords.append((x, y, z))
                types.append(2)              # wood
    for y in range(6):
        for z in range(1, 5):
            for x in [0, 5]:
                coords.append((x, y, z))
                types.append(2)              # wood
    # front door area at (2,0,1)-(3,0,2) – unknown to be predicted
    door_coords = [(2, 0, 1), (3, 0, 1), (2, 0, 2), (3, 0, 2)]
    for c in door_coords:
        coords.append(c)
        types.append(-1)                     # unknown
    coords = torch.tensor(coords, dtype=torch.float32)
    types = torch.tensor(types, dtype=torch.long)
    return coords, types

# --------------------------------------------------------------------------------
#  Entrypoint
# --------------------------------------------------------------------------------
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="data", help="3D‑Craft root")
    parser.add_argument("--batch_size", type=int, default=64)
    parser.add_argument("--epochs", type=int, default=10)
    parser.add_argument("--lr", type=float, default=1e-3)
    args = parser.parse_args()

    dm_train = Craft3DBlockTypeDataset(args.data_dir, "train")
    dm_val   = Craft3DBlockTypeDataset(args.data_dir, "val")

    train_loader = DataLoader(dm_train, batch_size=args.batch_size, shuffle=True, num_workers=4)
    val_loader   = DataLoader(dm_val,   batch_size=args.batch_size, shuffle=False, num_workers=4)

    module = BlockTypeModule(vars(args))

    trainer = L.Trainer(
        max_epochs=args.epochs,
        default_root_dir="logs",
        log_every_n_steps=1,
        enable_checkpointing=True,
        gradient_clip_val=1.0,
    )

    trainer.fit(module, train_loader, val_loader)

if __name__ == "__main__":
    main()
